import argparse

import arviz as az
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pymc as pm
import pymc_bart as pmb

RANDOM_SEED = 8457


def test_propensity(args):
    nhefs_df = pd.read_csv(pm.get_data("nhefs.csv"))

    X = nhefs_df.astype("float64").copy()
    y = nhefs_df["outcome"].astype("float64")
    t = nhefs_df["trt"].astype("float64")
    X = X.drop(["trt", "outcome"], axis=1)

    coords = {"coeffs": list(X.columns), "obs": range(len(X))}
    with pm.Model(coords=coords) as model_ps:
        X_data = pm.MutableData("X", X)
        t_data = pm.MutableData("t", t)

        mu = pmb.BART("mu", X, t, m=args.trees)
        p = pm.Deterministic("p", pm.math.invprobit(mu))

        t_pred = pm.Bernoulli("t_pred", p=p, observed=t_data, dims="obs")

        idata = pm.sample(
            tune=args.tune,
            draws=args.draws,
            step=[
                pmb.PGBART([mu], batch=tuple(args.batch), num_particles=args.particles)
            ],
            random_seed=RANDOM_SEED,
        )


def test_bikes(args):
    bikes = pd.read_csv(pm.get_data("bikes.csv"))
    X = bikes[["hour", "temperature", "humidity", "workingday"]]
    Y = bikes["count"]

    with pm.Model() as model_bikes:
        alpha = pm.Exponential("alpha", 1.0)
        mu = pmb.BART("mu", X, np.log(Y), m=args.trees)
        y = pm.NegativeBinomial("y", mu=pm.math.exp(mu), alpha=alpha, observed=Y)

        idata = pm.sample(
            tune=args.tune,
            draws=args.draws,
            step=[
                pmb.PGBART([mu], batch=tuple(args.batch), num_particles=args.particles)
            ],
            random_seed=RANDOM_SEED,
        )


def test_coal(args):
    coal = np.loadtxt(pm.get_data("coal.csv"))

    # discretize data
    years = int(coal.max() - coal.min())
    bins = years // 4
    hist, x_edges = np.histogram(coal, bins=bins)
    # compute the location of the centers of the discretized data
    x_centers = x_edges[:-1] + (x_edges[1] - x_edges[0]) / 2
    # xdata needs to be 2D for BART
    x_data = x_centers[:, None]
    # express data as the rate number of disaster per year
    y_data = hist

    with pm.Model() as model_coal:
        mu = pmb.BART(
            "mu",
            X=x_data,
            Y=np.log(y_data),
            m=args.trees,
            # split_rules=["ContinuousSplit"]
        )
        exp_mu = pm.Deterministic("exp_mu", pm.math.exp(mu))
        y_pred = pm.Poisson("y_pred", mu=exp_mu, observed=y_data)

        idata = pm.sample(
            tune=args.tune,
            draws=args.draws,
            chains=4,
            step=[
                pmb.PGBART([mu], batch=tuple(args.batch), num_particles=args.particles)
            ],
            random_seed=RANDOM_SEED,
        )
    #     step = pmb.PGBART([mu], batch=tuple(args.batch), num_particles=args.particles)

    # for i in range(1500):
    #     sum_trees, stats = step.astep(i)
    #     print(f"iter: {i}, time: {stats[0].get('time')}")

    _, ax = plt.subplots(figsize=(10, 6))
    rates = idata.posterior["exp_mu"] / 4
    rate_mean = rates.mean(dim=["draw", "chain"])
    ax.plot(x_centers, rate_mean, "w", lw=3)
    ax.plot(x_centers, y_data / 4, "k.")
    az.plot_hdi(x_centers, rates, smooth=False)
    az.plot_hdi(x_centers, rates, hdi_prob=0.5, smooth=False, plot_kwargs={"alpha": 0})
    ax.plot(coal, np.zeros_like(coal) - 0.5, "k|")
    ax.set_xlabel("years")
    ax.set_ylabel("rate")
    plt.show()


def main(args):

    if args.model == "coal":
        test_coal(args)
    elif args.model == "bikes":
        test_bikes(args)
    elif args.model == "propensity":
        test_propensity(args)
    else:
        raise TypeError("Invalid model argument passed.")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--model", type=str, default="coal", help="Model name")
    parser.add_argument("--trees", type=int, default=50, help="Number of trees")
    parser.add_argument("--particles", type=int, default=20, help="Number of particles")
    parser.add_argument("--tune", type=int, default=1000, help="Number of tuning steps")
    parser.add_argument("--draws", type=int, default=1000, help="Number of draws")
    parser.add_argument("--batch", nargs="+", default=(1.0, 1.0), type=float)
    args = parser.parse_args()
    main(args)



import ctypes
import warnings

import numpy as np
import numba
import pandas as pd
import pymc as pm
import pytensor

from pymc.pytensorf import (
    compile_pymc,
    inputvars,
    join_nonshared_inputs,
    make_shared_replacements
)
from numba import carray, cfunc, extending, float64, types, njit
from numba.core import cgutils


@numba.extending.intrinsic
def address_as_void_pointer(typingctx, src):
    """Returns a void pointer from a given memory address

    Full credit goes to: https://stackoverflow.com/a/61550054
    """

    sig = types.voidptr(src)

    def codegen(cgctx, builder, sig, args):
        return builder.inttoptr(args[0], cgutils.voidptr_t)

    return sig, codegen


class CompiledPyMCModel:
    def __init__(self, model, vars):
        n_dims, pytensor_logp_fn, shared = self._make_functions(model, vars)
        self.n_dims = n_dims
        self.logp_fn_ptr = pytensor_logp_fn
        self.shared = shared

        self.logp_args = self._make_persistent_arrays()
        self.logp = self._generate_logp_function()

    def get_function_pointer(self):
        return self.logp.address

    def _make_functions(self, model, vars):
        initial_values = model.initial_point()
        shape = initial_values.get(str(vars[0])).shape[0]
        shared = make_shared_replacements(initial_values, vars, model)
        out_vars = [model.datalogp]

        new_out, new_joined_inputs = join_nonshared_inputs(
            initial_values,
            out_vars,
            vars,
            shared
        )

        logp_fn = compile_pymc(
            inputs=[new_joined_inputs],
            outputs=new_out[0],
            mode="NUMBA"
        )
        logp_fn.trust_input = True

        return shape, logp_fn, shared

    def _make_persistent_arrays(self):
        arrays = [item.storage[0].copy() for item in self.logp_fn_ptr.input_storage[1:]]
        assert all(arr.dtype == np.float64 for arr in arrays)
        return arrays


    def update_shared_arrays(self):
        """Update the persistent shared arrays with new values from the function storage"""
        for array, storage in zip(self.logp_args, self.logp_fn_ptr.input_storage[1:]):
            new_arr = storage.storage[0]
            assert array.shape == new_arr.shape
            array *= 0.0
            array += new_arr

    def _generate_logp_function(self):
        logp_fn = self.logp_fn_ptr.vm.jit_fn
        shared_arrays = self.logp_args

        code = [
            "def _logp(ptr, size):",
            "    data = carray(ptr, (size, ), dtype=float64)"
        ]

        for i, array in enumerate(shared_arrays):
            line = "    arg{} = carray(address_as_void_pointer({}), {}, dtype={})".format(
                i, array.ctypes.data, array.shape, array.dtype
            )
            code.append(line)

        ret = f"    return logp_fn(data, {', '.join(f'arg{i}' for i in range(len(shared_arrays)))})[0].item()"
        code.append(ret)
        source = "\n".join(code)

        ldict = locals()
        gdict = {**globals(), **locals()}
        exec(source, gdict, ldict)

        logp = njit(ldict["_logp"])

        sig = types.float64(
            types.CPointer(types.float64),
            types.intc,
        )
        logp = cfunc(sig)(logp)

        return logp



#   Copyright 2022 The PyMC Developers
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from time import perf_counter
from typing import Optional, Tuple

import numpy as np

from pymc.initial_point import PointType
from pymc.model import Model, modelcontext
from pymc.pytensorf import inputvars
from pymc.step_methods.arraystep import ArrayStepShared
from pymc.step_methods.compound import Competence
from pytensor.graph.basic import Variable

from pymc_bart.bart import BARTRV
from pymc_bart.compile_pymc import CompiledPyMCModel
from pymc_bart.pymc_bart import initialize, step


class PGBART(ArrayStepShared):
    """
    Particle Gibss BART sampling step.

    Parameters
    ----------
    vars: list
        List of value variables for sampler
    num_particles : tuple
        Number of particles. Defaults to 10
    batch : tuple
        Number of trees fitted per step. The first element is the batch size during tuning and the
        second the batch size after tuning.  Defaults to  (0.1, 0.1), meaning 10% of the `m` trees
        during tuning and after tuning.
    model: PyMC Model
        Optional model for sampling step. Defaults to None (taken from context).
    """

    name = "pgbart"
    default_blocked = False
    generates_stats = True
    stats_dtypes_shapes: dict[str, tuple[type, list]] = {
        "variable_inclusion": (object, []),
        "tune": (bool, []),
        "time": (float, [])
    }

    def __init__(  # noqa: PLR0915
        self,
        vars=None,  # pylint: disable=redefined-builtin
        num_particles: int = 10,
        batch: Tuple[float, float] = (0.1, 0.1),
        model: Optional[Model] = None,
        initial_point: PointType | None = None,
        compile_kwargs: dict | None = None,  # pylint: disable=unused-argument
    ):

        model = modelcontext(model)
        if initial_point is None:
            initial_point = model.initial_point()

        # Get the instances of the BART random variable from the PyMC model
        if vars is None:
            vars = model.value_vars
        else:
            vars = [model.rvs_to_values.get(var, var) for var in vars]
            vars = inputvars(vars)
        value_bart = vars[0]
        self.bart = model.values_to_rvs[value_bart].owner.op

        if isinstance(self.bart.X, Variable):
            self.X = self.bart.X.eval()
        else:
            self.X = self.bart.X

        if isinstance(self.bart.Y, Variable):
            self.Y = self.bart.Y.eval()
        else:
            self.Y = self.bart.Y

        self.m = self.bart.m
        self.response = self.bart.response

        shape = initial_point[value_bart.name].shape
        self.shape = 1 if len(shape) == 1 else shape[0]

        # Set trees_shape (dim for separate tree structures)
        # and leaves_shape (dim for leaf node values)
        # One of the two is always one, the other equal to self.shape
        self.trees_shape = self.shape if self.bart.separate_trees else 1
        self.leaves_shape = self.shape if not self.bart.separate_trees else 1

        if self.bart.split_prior.size == 0:
            self.alpha_vec = np.ones(self.X.shape[1])
        else:
            self.alpha_vec = self.bart.split_prior

        if self.bart.split_rules:
            self.split_rules = self.bart.split_rules
        else:
            self.split_rules = ["ContinuousSplit"] * self.X.shape[1]

        # If data is binary
        self.leaf_sd = np.ones((self.trees_shape, self.leaves_shape))

        y_unique = np.unique(self.bart.Y)
        if y_unique.size == 2 and np.all(y_unique == [0, 1]):
            self.leaf_sd *= 3 / self.m**0.5
        else:
            self.leaf_sd *= self.bart.Y.std() / self.m ** 0.5

        # Compile the PyMC model to create a C callback. This function pointer is
        # passed to Rust and called using Rust's foreign function interface (FFI)
        self.compiled_pymc_model = CompiledPyMCModel(model, vars)

        # Initialize the Rust Particle-Gibbs sampler
        self.state = initialize(
            X=self.X,
            y=self.bart.Y,
            logp=self.compiled_pymc_model.get_function_pointer(),
            alpha=self.bart.alpha,
            beta=self.bart.beta,
            split_prior=self.alpha_vec,
            split_rules=self.split_rules,
            response=self.bart.response,
            n_trees=self.bart.m,
            n_particles=num_particles,
            leaf_sd=self.leaf_sd,
            batch=batch,
            _leaves_shape=self.leaves_shape,
        )

        self.tune = True
        super().__init__(vars, self.compiled_pymc_model.shared)

    def astep(self, _):
        # Record time to quantify performance improvements
        t0 = perf_counter()
        self.compiled_pymc_model.update_shared_arrays()
        sum_trees, variable_inclusion = step(self.state, self.tune)
        t1 = perf_counter()

        stats = {
            "variable_inclusion": variable_inclusion,
            "tune": self.tune,
            "time": t1 - t0
        }
        return sum_trees, [stats]

    @staticmethod
    def competence(var, has_grad):
        """PGBART is only suitable for BART distributions."""
        dist = getattr(var.owner, "op", None)
        if isinstance(dist, BARTRV):
            return Competence.IDEAL
        return Competence.INCOMPATIBLE



import pymc as pm

from pymc_bart.bart import BART
from pymc_bart.compile_pymc import CompiledPyMCModel
from pymc_bart.pgbart import PGBART
from pymc_bart.utils import (
    compute_variable_importance,
    plot_convergence,
    plot_dependence,
    plot_ice,
    plot_pdp,
    plot_scatter_submodels,
    plot_variable_importance,
    plot_variable_inclusion,
)

__all__ = [
    "BART",
    "PGBART",
    "CompiledPyMCModel",
    "compute_variable_importance",
    "plot_convergence",
    "plot_dependence",
    "plot_ice",
    "plot_pdp",
    "plot_scatter_submodels",
    "plot_variable_importance",
    "plot_variable_inclusion",
]

pm.STEP_METHODS = list(pm.STEP_METHODS) + [PGBART]
